#!/usr/bin/perl
use strict;

sub clean {
	my @clean;
	push(@clean,$_[0]);
	foreach my $this (@_) {
		#0: skip, change
		#1: push
		my $check = 1;
		for (my $i=0;$i<@clean;$i++) {
			my $that = $clean[$i];
			if ($that =~ /$this/) {
				$check = 0;
				last;
			}
			elsif ($this =~ /$that/) {
				$check = 0;
				$clean[$i] = $this;
				last;
			}
		}
		if ($check == 1) {
			push(@clean,$this);
		}
	}
	@clean = sort{length($b) <=> length($a)} @clean;
	return @clean;
}

sub uniq {
	my %seen;
	map $seen{$_}=1,@_;
	my @uniq = sort{$a cmp $b} (keys %seen);
	return @uniq;
}


if (!(-e './RemoveRfam/allseq.fa')) {
	die('Complete step 2 first!');
}

#my $trinity = '/home/jsyf/tool/trinity/trinityrnaseq_r20140717/Trinity';
my $trinity = '';
open(IN,'<program_list');
while (<IN>) {
	chomp;
	if ($_ eq '') {
		next;
	}

	my ($program,$location) = split(/[\t=]/,$_);

	$program =~ s/\s+//g;
	$location  =~ s/\s+//g;

	if ($program eq 'Trinity') {
		$trinity = $location;
	}
}
close(IN);

if (($trinity eq '') || !(-e $trinity)) {
	die("Install trinity and edit program_list!\n");
}

system('echo "Step 3 Start '.time().'" >> Process_Log');

system($trinity.' --bypass_java_version_check --KMER_SIZE=15 --seqType fa --max_memory 10G --single allseq.fa --SS_lib_type F --CPU 6 --min_contig_length 20 --output Trinity');
chdir('Trinity');
mkdir('bowtie');
chdir('bowtie');

system('echo "Step 3: Trinity done" >> ../../Process_Log');

system('cp -s ../../RemoveRfam/allseq.fa ./');
system('cp -s ../inchworm.K15.L15.fa ./contig.fa');

system('bowtie-build -f contig.fa contig');
system('bowtie -f --norc -v 0 -a contig allseq.fa allseq_contig');

system('echo "Step 3: Mapping contigs done" >> ../../Process_Log');

my %read_seq;
open(IN,'<allseq.fa');
while (my $name = <IN>) {
	chomp($name);

	chomp(my $seq = <IN>);

	$name = substr($name,1);
	$read_seq{$name} = $seq;
}
close(IN);

my %contig_seq;
my $name;
my $seq;
open(IN,'<contig.fa');
while (<IN>) {
	chomp;
	if ($_ eq '') {
		next;
	}

	if ($_ =~ /^>/) {
		if ($name) {
			$contig_seq{$name} = $seq;
			$name = '';
			$seq = '';
		}
		$name = $_;
		($name) = split(' ',$name);
		$name = substr($name,1);
	}
	else {
		$seq .= $_;
	}
}
if ($name) {
	$contig_seq{$name} = $seq;
}
close(IN);
undef($name);
undef($seq);


my %collect;
open(IN,'<allseq_contig');
while (<IN>) {
	chomp;
	if ($_ eq '') {
		next;
	}

	my ($read,$strand,$contig,$site,$seq) = split("\t",$_);

	my $start = $site - 5;
	if ($start < 0) {
		$start = 0;
	}

	my $len = length($seq);

	push(@{$collect{$read}},join(',',substr($contig_seq{$contig},$start,($site - $start)),substr($contig_seq{$contig},$site,$len),substr($contig_seq{$contig},($site + $len),5)));
}
close(IN);

open(OUT,'>full_seq');
foreach my $read (sort{$a cmp $b} (keys %collect)) {
	my @all_pattern = @{$collect{$read}};
	@all_pattern = &uniq(@all_pattern);
	@all_pattern = &clean(@all_pattern);
#       if (@all_pattern > 1) {
		foreach my $row (@all_pattern) {
			$row =~ s/,/\t/g;
			print OUT join("\t",$read,$row)."\n";
		}
#       }
}
close(OUT);

system('echo "Step 3: Extract flanking sequence done" >> ../../Process_Log');

chdir('../../');
system('cp -s ./Trinity/bowtie/full_seq ./');

system('echo "Step 3 done '.time().'" >> Process_Log');
