#!/usr/bin/perl
use strict;

sub checkpattern {
	#return value:
	#	0: OK
	#	1: branched pairing
	#	2: no 3'-overhang (left)
	#	3: no 3'-overhang (right)
	#	4: too many mismatch
	#	5: no match
	my @sequence = @{\@_}[0..1];
	my @pattern = @{\@_}[2..3];

	if (($pattern[0] =~ /\)/) || ($pattern[1] =~ /\(/)) {
		return 1;
	}
	elsif (($pattern[0] !~ /\(/) || ($pattern[1] !~ /\)/)) {
		return 5;
	}
	elsif (($pattern[0] =~ /^\.*\(\.*$/) || ($pattern[1] =~ /^\.*\)\.*$/)) {
		return 5;
	}
	else {
		my @CheckMis_pattern = @pattern;
		$CheckMis_pattern[1] = reverse($CheckMis_pattern[1]);
		$CheckMis_pattern[1] =~ tr/()/)(/;

		my @CheckMis_separate_pattern;
		for (my $i=0;$i<2;$i++) {
			$CheckMis_separate_pattern[$i] = [$CheckMis_pattern[$i] =~ /^(\.*)([\(\)].*[\(\)])(\.*)$/];
		}

		my @CheckMis_count_array;
		for (my $i=0;$i<2;$i++) {
			$CheckMis_count_array[$i] = [split(/[\(\)]/,$CheckMis_separate_pattern[$i][1],-1)];
		}

		my $CheckMis_n = @{$CheckMis_count_array[0]};
		$CheckMis_n = $CheckMis_n - 1;

		my $CheckMis_count = 0;
		for (my $i=1;$i<$CheckMis_n;$i++) {
			my @CheckMis_mismatch;
			for (my $j=0;$j<2;$j++) {
				$CheckMis_mismatch[$j] = length($CheckMis_count_array[$j][$i]);
			}

			if ($CheckMis_mismatch[0] > $CheckMis_mismatch[1]) {
				$CheckMis_count += $CheckMis_mismatch[0];
			}
			else {
				$CheckMis_count += $CheckMis_mismatch[1];
			}
		}

		my %CheckMis_overhang;
		foreach my $ind (0,2) {
			my $side = '';
			if ($ind == 0) {
				$side = 'left';
			}
			elsif ($ind == 2) {
				$side = 'right';
			}
	
			my $cmp = length($CheckMis_separate_pattern[0][$ind]) <=> length($CheckMis_separate_pattern[1][$ind]);
			my $val = length($CheckMis_separate_pattern[0][$ind]) - length($CheckMis_separate_pattern[1][$ind]);
			$val = $val * $cmp;

			$CheckMis_overhang{$side} = [$cmp,$val];

			if ($cmp == 1) {
				$CheckMis_count += length($CheckMis_separate_pattern[1][$ind]);
			}
			elsif ($cmp == -1) {
				$CheckMis_count += length($CheckMis_separate_pattern[0][$ind]);
			}
			else {
				$CheckMis_count += length($CheckMis_separate_pattern[1][$ind]);
			}
		}

		if (($CheckMis_overhang{'left'}[0] == 1) || ($CheckMis_overhang{'left'}[1] > 5)) {
			return 2;
		}
		elsif (($CheckMis_overhang{'right'}[0] == -1) || ($CheckMis_overhang{'right'}[1] > 5)) {
			return 3;
		}
		elsif ($CheckMis_count > 5) {
			return 4;
		}
		else {
			return 0;
		}
	}
}

my $cofold = '';
open(IN,'<../program_list');
while (<IN>) {
        chomp;
        if ($_ eq '') {
                next;
        }

        my ($program,$location) = split(/[\t=]/,$_);

	$program =~ s/\s+//g;
	$location =~ s/\s+//g;

        if ($program eq 'RNAcofold') {
                $cofold = $location;
        }
}
close(IN);

if (($cofold eq '') || !(-e $cofold)) {
        die("RNAcofold is not exist!\nPlease edit program_list!\n");
}


#if (!(-e 'Pair')) {
#        die('Complete step 4.1 first!'.$/);
#}
#elsif (!(-d 'Pair')) {
#	die('Complete step 4.1 first!'.$/);
#}
#chdir('Pair');

my @seq_guide;
open(IN,'<Cofold_use_guide.fa');
while (my $name = <IN>) {
	chomp($name);
	$name = substr($name,1);
	chomp(my $seq = <IN>);

	push(@seq_guide,[$name,$seq]);
}
close(IN);

my %fullseq;
open(IN,'<../full_seq');
while (<IN>) {
	chomp;
	if ($_ eq '') {
		next;
	}

	my @array = split("\t",$_);

	push(@{$fullseq{$array[0]}},$array[1].$array[2].$array[3]);
}
close(IN);

my $file = $ARGV[0];
my $folder = '';
if ($file =~ /Cofold_use_passenger_(\d+).fa/) {
	$folder = 'working_'.$1;
}
else {
	system('echo "Step 4.2: Incorrect passenger sequence file" >> ../Process_Log');
	exit;
}

my $njobs = $ARGV[1];
if ($njobs =~ /\D/) {
	system('echo "Step 4.2: Incorrect thread number." >> ../Process_Log');
	exit;
}
$njobs = 10000 / $njobs;
if ($njobs != 1) {
	$njobs = int($njobs);
	$njobs = $njobs + 1;
}

my @passenger_group;
my $group_tmp = 0;
my $count_tmp = 0;
open(IN,'<'.$file);
while (my $name = <IN>) {
	chomp($name);
	chomp(my $seq = <IN>);

	$name = substr($name,1);

	$count_tmp ++;
	if ($count_tmp > $njobs) {
		$count_tmp -= $njobs;
		$group_tmp ++;
	}

	push(@{$passenger_group[$group_tmp]},[$name,$seq]);
}
close(IN);

my $n_group = @passenger_group;

chomp(my $pwd = `pwd`);
mkdir($folder);
chdir($folder);

for (my $group_ref=0;$group_ref<$n_group;$group_ref++) {
	if (!defined($passenger_group[$group_ref]) || (@{$passenger_group[$group_ref]} == 0)) {
		next;
	}

	if (fork() == 0) {
		my $full_in = 'tmp_'.$group_ref.'_full.fa';
		my $read_in = 'tmp_'.$group_ref.'_read.fa';

		my $full_out = 'tmp_'.$group_ref.'_full.cofold';
		my $read_out = 'tmp_'.$group_ref.'_read.cofold';

		my $full_filtered = 'tmp_'.$group_ref.'_full.filtered';
		my $read_filtered = 'tmp_'.$group_ref.'_read.filtered';

		if (!(-e $full_in) || !(-e $read_in)) {
			open(FULL,'>'.$full_in);
			open(OUT,'>'.$read_in);
			foreach my $info_g (@seq_guide) {
				my ($name_g,$seq_g) = @{$info_g};

				foreach my $info_p (@{$passenger_group[$group_ref]}) {
					my ($name_p,$seq_p) = @{$info_p};

					print OUT '>'.$name_g.'_'.$name_p."\n".$seq_g.'&'.$seq_p."\n";

					my @full_g;
					if (defined($fullseq{$name_g}) && (@{$fullseq{$name_g}} > 0)) {
						@full_g = @{$fullseq{$name_g}};
					}
					else {
						@full_g = ($seq_g);
					}

					my @full_p;
					if (defined($fullseq{$name_p}) && (@{$fullseq{$name_p}} > 0)) {
						@full_p = @{$fullseq{$name_p}};
					}
					else {
						@full_p = ($seq_p);
					}

					for (my $i=0;$i<@full_g;$i++) {
						for (my $j=0;$j<@full_p;$j++) {
							print FULL '>'.$name_g.'_'.$name_p.'_'.$i.$j."\n".$full_g[$i].'&'.$full_p[$j]."\n";
						}
					}
				}
			}
			close(OUT);
			close(FULL);
		}

		if (!(-e $full_out) || !(-e $read_out)) {
			foreach my $row ([$full_in,$full_out],[$read_in,$read_out]) {
				my ($input,$output) = @{$row};

				my $n_in = `wc -l $input`;
				chomp($n_in);
				($n_in) = $n_in =~ /^(\d+)\s/g;

				#system('/home/u00wcc02/u00wcc02/jsyf_use/tool/ViennaRNA/RNAcofold --noLP --noPS < '.$input.' > '.$output);
				system($cofold.' --noLP --noPS < '.$input.' > '.$output);

				while ($? == 256) {
					my $n_out = `grep -c '^>' $output`;

					chomp($n_out);

					my $n_retry = $n_in - ($n_out*2);
					if ($n_in < $n_out) { last; }

					#system('tail -n '.$n_retry.' '.$input.' | /home/u00wcc02/u00wcc02/jsyf_use/tool/ViennaRNA/RNAcofold --noLP --noPS >> '.$output);
					system('tail -n '.$n_retry.' '.$input.' | '.$cofold.' --noLP --noPS >> '.$output);
				}

				system('echo "Step 4.2: Generate '.$output.' done!" >> ../../Process_Log');
			}
		}

		system('grep -c "^>" '.join(' ',$full_in,$full_out,$read_in,$read_out))."\n";

		system('rm '.$full_in);
		system('rm '.$read_in);

		my %pair_structure;
		foreach my $cofold ('read','full') {
			my @name = ();
			my $full_id = '';
			my @seq = ();
			my @pattern = ();
			my $energy = 0;
			open(IN,'<tmp_'.$group_ref.'_'.$cofold.'.cofold');
			while (<IN>) {
				chomp;
				if ($_ eq '') {
					next;
				}

				if ($_ =~ /^>/) {
					if (@name == 2) {
						if ((@seq == 2) && (@pattern == 2)) {
							if (($cofold eq 'read') || (defined($pair_structure{$name[0]}{$name[1]}{'read'}))) {
								if (&checkpattern(@seq,@pattern) == 0) {
									if ($cofold eq 'read') {
										$pair_structure{$name[0]}{$name[1]}{$cofold} = [@seq,@pattern,$energy];
									}
									elsif ($cofold eq 'full') {
										$pair_structure{$name[0]}{$name[1]}{$cofold}{$full_id} = [@seq,@pattern,$energy];
									}
								}
							}
						}
					}
					@name = ();
					$full_id = '';
					@seq = ();
					@pattern = ();
					$energy = 0;
					if ($cofold eq 'read') {
						if ($_ =~ />(allseq\d+_rich\d+)_(allseq\d+_rich\d+)/) {
							@name = ($1,$2);
						}
					}
					elsif ($cofold eq 'full') {
						if ($_ =~ />(allseq\d+_rich\d+)_(allseq\d+_rich\d+)_(\d{2})/) {
							@name = ($1,$2);
							$full_id = $3;
						}
					}
				}
				elsif ($_ =~ / +\(/) {
					@pattern = split('&',$`);
					$energy = $';
					$energy =~ s/[\(\)\s]//g;
				}
				else {
					my $seq_tmp = $_;
					$seq_tmp =~ s/U/T/g;
					@seq = split('&',$seq_tmp);
				}
			}
			close(IN);
		}

		open(OUTR,'>'.$read_filtered);
		open(OUTF,'>'.$full_filtered);
		foreach my $name_a (sort{$a cmp $b} (keys %pair_structure)) {
			foreach my $name_b (sort{$a cmp $b} (keys %{$pair_structure{$name_a}})) {
				if (!defined($pair_structure{$name_a}{$name_b}{'read'}) || (@{$pair_structure{$name_a}{$name_b}{'read'}} == 0)) {
					next;
				}

				print OUTR join("\t",@{$pair_structure{$name_a}{$name_b}{'read'}})."\n";

				if (!defined($pair_structure{$name_a}{$name_b}{'full'}) || ((keys %{$pair_structure{$name_a}{$name_b}{'full'}}) == 0)) {
					next;
				}

				foreach my $full_id (sort{$a cmp $b} (keys %{$pair_structure{$name_a}{$name_b}{'full'}})) {
					print OUTF join("\t",@{$pair_structure{$name_a}{$name_b}{'full'}{$full_id}})."\n";
				}
			}
		}
		close(OUTR);
		close(OUTF);

		system('rm '.$full_out);
		system('rm '.$read_out);

		system('echo "Step 4.2: Group '.$group_ref.' done!" >> ../../Process_Log');
		exit;
	}
}

wait for 0 .. $n_group;

$folder =~ /^working_(\d+)$/;
my $file_id = $1;

foreach my $done_type ('full','read') {
	my $done_file = 'Cofold_filtered_'.$done_type.'_'.$file_id.'.txt';

	system('touch '.$done_file);
	system('rm -f '.$done_file);
	system('touch '.$done_file);

	for (my $group_ref=0;$group_ref<$n_group;$group_ref++) {
		system('cat tmp_'.$group_ref.'_'.$done_type.'.filtered >> '.$done_file);
		system('rm -f tmp_'.$group_ref.'_'.$done_type.'.filtered');
	}
}

system('echo "File: '.$file.' done!" >> ../../Process_Log');

system('touch '.$file.'.finish');
#}
chdir($pwd);
